"""Wide_Eastasian table. Created by setup.py."""
# Generated: 2016-07-02T04:20:28.048222
# Source: EastAsianWidth-9.0.0.txt
# Date:  2016-05-27, 17:00:00 GMT [KW, LI]
WIDE_EASTASIAN = (
    (0x1100, 0x115f,),  # Hangul Choseong Kiyeok  ..Hangul Choseong Filler
    (0x231a, 0x231b,),  # Watch                   ..Hourglass
    (0x2329, 0x232a,),  # Left-pointing Angle Brac..Right-pointing Angle Bra
    (0x23e9, 0x23ec,),  # Black Right-pointing Dou..Black Down-pointing Doub
    (0x23f0, 0x23f0,),  # Alarm Clock             ..Alarm Clock
    (0x23f3, 0x23f3,),  # Hourglass With Flowing S..Hourglass With Flowing S
    (0x25fd, 0x25fe,),  # White Medium Small Squar..Black Medium Small Squar
    (0x2614, 0x2615,),  # Umbrella With Rain Drops..Hot Beverage
    (0x2648, 0x2653,),  # Aries                   ..Pisces
    (0x267f, 0x267f,),  # Wheelchair Symbol       ..Wheelchair Symbol
    (0x2693, 0x2693,),  # Anchor                  ..Anchor
    (0x26a1, 0x26a1,),  # High Voltage Sign       ..High Voltage Sign
    (0x26aa, 0x26ab,),  # Medium White Circle     ..Medium Black Circle
    (0x26bd, 0x26be,),  # Soccer Ball             ..Baseball
    (0x26c4, 0x26c5,),  # Snowman Without Snow    ..Sun Behind Cloud
    (0x26ce, 0x26ce,),  # Ophiuchus               ..Ophiuchus
    (0x26d4, 0x26d4,),  # No Entry                ..No Entry
    (0x26ea, 0x26ea,),  # Church                  ..Church
    (0x26f2, 0x26f3,),  # Fountain                ..Flag In Hole
    (0x26f5, 0x26f5,),  # Sailboat                ..Sailboat
    (0x26fa, 0x26fa,),  # Tent                    ..Tent
    (0x26fd, 0x26fd,),  # Fuel Pump               ..Fuel Pump
    (0x2705, 0x2705,),  # White Heavy Check Mark  ..White Heavy Check Mark
    (0x270a, 0x270b,),  # Raised Fist             ..Raised Hand
    (0x2728, 0x2728,),  # Sparkles                ..Sparkles
    (0x274c, 0x274c,),  # Cross Mark              ..Cross Mark
    (0x274e, 0x274e,),  # Negative Squared Cross M..Negative Squared Cross M
    (0x2753, 0x2755,),  # Black Question Mark Orna..White Exclamation Mark O
    (0x2757, 0x2757,),  # Heavy Exclamation Mark S..Heavy Exclamation Mark S
    (0x2795, 0x2797,),  # Heavy Plus Sign         ..Heavy Division Sign
    (0x27b0, 0x27b0,),  # Curly Loop              ..Curly Loop
    (0x27bf, 0x27bf,),  # Double Curly Loop       ..Double Curly Loop
    (0x2b1b, 0x2b1c,),  # Black Large Square      ..White Large Square
    (0x2b50, 0x2b50,),  # White Medium Star       ..White Medium Star
    (0x2b55, 0x2b55,),  # Heavy Large Circle      ..Heavy Large Circle
    (0x2e80, 0x2e99,),  # Cjk Radical Repeat      ..Cjk Radical Rap
    (0x2e9b, 0x2ef3,),  # Cjk Radical Choke       ..Cjk Radical C-simplified
    (0x2f00, 0x2fd5,),  # Kangxi Radical One      ..Kangxi Radical Flute
    (0x2ff0, 0x2ffb,),  # Ideographic Description ..Ideographic Description
    (0x3000, 0x303e,),  # Ideographic Space       ..Ideographic Variation In
    (0x3041, 0x3096,),  # Hiragana Letter Small A ..Hiragana Letter Small Ke
    (0x3099, 0x30ff,),  # Combining Katakana-hirag..Katakana Digraph Koto
    (0x3105, 0x312d,),  # Bopomofo Letter B       ..Bopomofo Letter Ih
    (0x3131, 0x318e,),  # Hangul Letter Kiyeok    ..Hangul Letter Araeae
    (0x3190, 0x31ba,),  # Ideographic Annotation L..Bopomofo Letter Zy
    (0x31c0, 0x31e3,),  # Cjk Stroke T            ..Cjk Stroke Q
    (0x31f0, 0x321e,),  # Katakana Letter Small Ku..Parenthesized Korean Cha
    (0x3220, 0x3247,),  # Parenthesized Ideograph ..Circled Ideograph Koto
    (0x3250, 0x32fe,),  # Partnership Sign        ..Circled Katakana Wo
    (0x3300, 0x4dbf,),  # Square Apaato           ..
    (0x4e00, 0xa48c,),  # Cjk Unified Ideograph-4e..Yi Syllable Yyr
    (0xa490, 0xa4c6,),  # Yi Radical Qot          ..Yi Radical Ke
    (0xa960, 0xa97c,),  # Hangul Choseong Tikeut-m..Hangul Choseong Ssangyeo
    (0xac00, 0xd7a3,),  # Hangul Syllable Ga      ..Hangul Syllable Hih
    (0xf900, 0xfaff,),  # Cjk Compatibility Ideogr..
    (0xfe10, 0xfe19,),  # Presentation Form For Ve..Presentation Form For Ve
    (0xfe30, 0xfe52,),  # Presentation Form For Ve..Small Full Stop
    (0xfe54, 0xfe66,),  # Small Semicolon         ..Small Equals Sign
    (0xfe68, 0xfe6b,),  # Small Reverse Solidus   ..Small Commercial At
    (0xff01, 0xff60,),  # Fullwidth Exclamation Ma..Fullwidth Right White Pa
    (0xffe0, 0xffe6,),  # Fullwidth Cent Sign     ..Fullwidth Won Sign
    (0x16fe0, 0x16fe0,),  # (nil)                   ..
    (0x17000, 0x187ec,),  # (nil)                   ..
    (0x18800, 0x18af2,),  # (nil)                   ..
    (0x1b000, 0x1b001,),  # Katakana Letter Archaic ..Hiragana Letter Archaic
    (0x1f004, 0x1f004,),  # Mahjong Tile Red Dragon ..Mahjong Tile Red Dragon
    (0x1f0cf, 0x1f0cf,),  # Playing Card Black Joker..Playing Card Black Joker
    (0x1f18e, 0x1f18e,),  # Negative Squared Ab     ..Negative Squared Ab
    (0x1f191, 0x1f19a,),  # Squared Cl              ..Squared Vs
    (0x1f200, 0x1f202,),  # Square Hiragana Hoka    ..Squared Katakana Sa
    (0x1f210, 0x1f23b,),  # Squared Cjk Unified Ideo..
    (0x1f240, 0x1f248,),  # Tortoise Shell Bracketed..Tortoise Shell Bracketed
    (0x1f250, 0x1f251,),  # Circled Ideograph Advant..Circled Ideograph Accept
    (0x1f300, 0x1f320,),  # Cyclone                 ..Shooting Star
    (0x1f32d, 0x1f335,),  # Hot Dog                 ..Cactus
    (0x1f337, 0x1f37c,),  # Tulip                   ..Baby Bottle
    (0x1f37e, 0x1f393,),  # Bottle With Popping Cork..Graduation Cap
    (0x1f3a0, 0x1f3ca,),  # Carousel Horse          ..Swimmer
    (0x1f3cf, 0x1f3d3,),  # Cricket Bat And Ball    ..Table Tennis Paddle And
    (0x1f3e0, 0x1f3f0,),  # House Building          ..European Castle
    (0x1f3f4, 0x1f3f4,),  # Waving Black Flag       ..Waving Black Flag
    (0x1f3f8, 0x1f43e,),  # Badminton Racquet And Sh..Paw Prints
    (0x1f440, 0x1f440,),  # Eyes                    ..Eyes
    (0x1f442, 0x1f4fc,),  # Ear                     ..Videocassette
    (0x1f4ff, 0x1f53d,),  # Prayer Beads            ..Down-pointing Small Red
    (0x1f54b, 0x1f54e,),  # Kaaba                   ..Menorah With Nine Branch
    (0x1f550, 0x1f567,),  # Clock Face One Oclock   ..Clock Face Twelve-thirty
    (0x1f57a, 0x1f57a,),  # (nil)                   ..
    (0x1f595, 0x1f596,),  # Reversed Hand With Middl..Raised Hand With Part Be
    (0x1f5a4, 0x1f5a4,),  # (nil)                   ..
    (0x1f5fb, 0x1f64f,),  # Mount Fuji              ..Person With Folded Hands
    (0x1f680, 0x1f6c5,),  # Rocket                  ..Left Luggage
    (0x1f6cc, 0x1f6cc,),  # Sleeping Accommodation  ..Sleeping Accommodation
    (0x1f6d0, 0x1f6d2,),  # Place Of Worship        ..
    (0x1f6eb, 0x1f6ec,),  # Airplane Departure      ..Airplane Arriving
    (0x1f6f4, 0x1f6f6,),  # (nil)                   ..
    (0x1f910, 0x1f91e,),  # Zipper-mouth Face       ..
    (0x1f920, 0x1f927,),  # (nil)                   ..
    (0x1f930, 0x1f930,),  # (nil)                   ..
    (0x1f933, 0x1f93e,),  # (nil)                   ..
    (0x1f940, 0x1f94b,),  # (nil)                   ..
    (0x1f950, 0x1f95e,),  # (nil)                   ..
    (0x1f980, 0x1f991,),  # Crab                    ..
    (0x1f9c0, 0x1f9c0,),  # Cheese Wedge            ..Cheese Wedge
    (0x20000, 0x2fffd,),  # Cjk Unified Ideograph-20..
    (0x30000, 0x3fffd,),  # (nil)                   ..
)
